import os
from concurrent.futures import (
    Executor, ProcessPoolExecutor, ThreadPoolExecutor, as_completed
)
from typing import Literal

from tqdm import tqdm


def parallel(
    func,
    params_list,
    executor_type: Executor,
    add_results: Literal['append', 'extend'] = 'append',
    disable_pbar: bool = False
):

    num_iter = len(params_list)

    if executor_type == ThreadPoolExecutor:
        max_workers = int(os.getenv('SF_NUM_THREADS', 100))

    elif executor_type == ProcessPoolExecutor:
        max_workers = int(os.getenv('SF_NUM_PROCESSORS', 8))

    if max_workers > num_iter:
        max_workers = num_iter

    with executor_type(max_workers=max_workers) as executor:
        futures = [executor.submit(func, **p) for p in params_list]
        results = []
        with tqdm(total=num_iter, disable=disable_pbar) as pbar:
            for future in as_completed(futures):
                if add_results == 'append':
                    results.append(future.result())
                elif add_results == 'extend':
                    results.extend(future.result())
                pbar.update()

        return results



import time
from contextlib import contextmanager

import pandas as pd


def timer(func):

    def wrapper(*args, **kwargs):
        s = time.time()
        result = func(*args, **kwargs)
        e = time.time() - s
        print(f'[{func.__name__}] {e:.2f}s')

        return result

    return wrapper


@contextmanager
def time_block(task_name: str):
    s = time.time()
    yield
    e = time.time() - s
    print(f"[{task_name}] {e:.2f}s")


def as_df(func) -> pd.DataFrame:
    '''Decorator to convert func output into a pd.DataFrame if output is 
    iterable.'''

    def wrapper(*args, **kwargs):
        objs = func(*args, **kwargs)
        if hasattr(objs, '__iter__'):
            return pd.DataFrame(objs)
        else:
            return objs

    return wrapper



from collections import namedtuple

import pandas as pd
from dynaconf.utils.boxing import DynaBox
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Engine

from synfras.config import config


def connect(urls: dict | DynaBox) -> namedtuple:
    return namedtuple('Database', urls.keys())(
        *[create_engine(url) for url in urls.values()]
    )


def query(
    engine: Engine,
    stmt: str,
    params: dict | list[dict] = None,
    fetch: bool = True,
    as_dict: bool = False,
    index_col: str | list[str] | None = None,
) -> None | list[dict] | pd.DataFrame:
    stmt = text(stmt)
    if fetch:
        result = pd.read_sql(
            stmt, engine, params=params, index_col=index_col, coerce_float=True
        )
        if config.get('timezone'):
            _handle_tz(result)
        if as_dict:
            result = result.to_dict(orient='records')
            result = _handle_length(result)
        return result
    else:
        with engine.connect() as conn:
            conn.execute(stmt, params)
            conn.commit()


def _handle_tz(result: pd.DataFrame) -> None:
    for col in result.columns:
        if isinstance(result[col].dtype, pd.DatetimeTZDtype):
            result[col] = result[col].dt.tz_convert(tz=config.timezone)


def _handle_length(result: list[dict]) -> list[dict] | dict:
    if len(result) == 1:
        return result[0]
    elif len(result) == 0:
        return {}
    return result
